#include "Arduino.h"
#include "Wire.h"
#include "EEPROM.h"

/*
   Pins - Arduino Uno - A4 (SDA), A5 (SCL) other boards see https://www.arduino.cc/en/Reference/Wire
          Arduino Nano - A4 (SCL), A5 (SDA) pins swaped from Uno.

   CatGenie 120 wire colors. This is just a starting point as they like to swap the colors. If this doesn't work you will need to probe them on your device.

   White - SDA - pin A4
   Black - SDL - pin A5
   Red - Power(5v)
   Blue - Ground

*/

byte I2C_ADDRESS = 80; // 80(7 bits) with a R/W bit appended

byte lastCount = -1;
int flash = 0;

byte* readRegister = new byte[0x10];
int readRegisterSize;
byte remaining = 0x78;

byte* response;
int size = 4;

// Data that will be "read" from RFID tag by block id
byte BLOCK_0[] = { 0x04, 0x00, 0x00, 0x00, 0x00 };
byte BLOCK_1[] = { 0x04, 0x78, 0x00, 0x00, 0x00 };
byte BLOCK_2[] = { 0x04, 0x01, 0x00, 0x78, 0x00 };
byte BLOCK_3[] = { 0x04, 0x04, 0x0F, 0x00, 0x00 };
byte BLOCK_4[] = { 0x04, 0x1F, 0x7E, 0x30, 0x2A };
byte BLOCK_5[] = { 0x04, 0x04, 0x0F, 0x2B, 0xFC };

byte* BLOCKS[] = {
  BLOCK_0, BLOCK_0, BLOCK_0, BLOCK_0,
  BLOCK_0, BLOCK_1, BLOCK_1, BLOCK_0,
  BLOCK_2, BLOCK_2, BLOCK_2, BLOCK_3,
  BLOCK_3, BLOCK_4, BLOCK_0, BLOCK_5
};

int InPin = PD2; // reset counter to 120 when grounded.

byte NODE_ID_RESPONSE[] = { 0x01, 0x3C };
byte UID_RESPONSE[] = { 0x08, 0xA3, 0x6E, 0x6A, 0x73, 0x09, 0x33, 0x02, 0xD0 };

void receiveEvent(int length) {
  // skip single byte write as its setting up the next read
  // and reading the ack breaks the following read
  if (length == 1) {
    // consume data but ignore
    Wire.read();
    return;
  }

  // all other instructions set the read register for a subsiquent block read (handled by requestEvent)
  for (int i = 0 ; Wire.available() ; i++) {
    readRegister[i] = Wire.read();
  }
  readRegisterSize = length;

  int command = readRegister[2] & 0x0f;

  switch (command) // RFID Tag command
  {
    case 0x06: // INITIATE, PCALL16, SLOT_MARKER
      response = NODE_ID_RESPONSE;
      size = 2;
      break;
    case 0x08: // RESET_BLOCK
      response = BLOCKS[readRegister[3]];
      size = 5;
      break;
    case 0x09: // WRITE_BLOCK
      response = BLOCKS[readRegister[3]];
      response[1] = readRegister[4];
      response[2] = readRegister[5];
      response[3] = readRegister[6];
      response[4] = readRegister[7];
      size = 0;
      break;
    case 0x0A: // AUTHENTICATE (ignored)
      size = 0;
      break;
    case 0x0B: // GET_UID
      response = UID_RESPONSE;
      size = 9;
      break;
    case 0x0C: // RESET_TO_INVENTORY
      size = 0;
      break;
    case 0x0E: // SELECT
      response = NODE_ID_RESPONSE;
      size = 2;
      break;
    case 0x0F: // COMPLETION
      size = 0;
      break;
  }
}

// WARNING - This needs to be pretty fast or we will miss data. Serial prints are too slow.
void requestEvent() {
  Wire.write(response, size);
}

void setup() {

  byte hdr1 = EEPROM.read(0);
  byte hdr2 = EEPROM.read(1);

  byte cnt1 = EEPROM.read(2);
  byte cnt2 = EEPROM.read(3);

  if (hdr1 != 'I' || hdr2 != 'P' || cnt1 != cnt2 || (cnt1 == 0 && cnt2 == 0))
  {
    EEPROM.write(0, 'I');
    EEPROM.write(1, 'P');
    cnt1 = 0x78;
    cnt2 = 0x78;

    updateCount(cnt1);
  }

  BLOCK_1[1] = cnt1;

  pinMode(InPin, INPUT_PULLUP);    // sets the digital pin 7 as input
  pinMode(LED_BUILTIN, OUTPUT);    // initialize digital pin LED_BUILTIN as an output.

  Wire.begin(I2C_ADDRESS);

  Wire.onRequest(requestEvent);
  Wire.onReceive(receiveEvent);
}

void updateCount(byte Count)
{
  EEPROM.write(2, Count);
  EEPROM.write(3, Count);
  lastCount = Count;
}

void loop() {
  delay(100);
  if (lastCount != BLOCK_1[1])
  {
    updateCount(BLOCK_1[1]);
  }

  if (flash > 0)
  {
    if (((flash >> 3) % 2) > 0)
    {
      digitalWrite(LED_BUILTIN, HIGH);   // turn the LED on (HIGH is the voltage level)
    }
    else
    {
      digitalWrite(LED_BUILTIN, LOW);    // turn the LED off by making the voltage LOW
    }
    flash--;
  }
  else
  {
    if (digitalRead(InPin) == 0)
    {
      BLOCK_1[1] = 0x78;
      BLOCK_2[3] = 0x78;
      updateCount(120);
      flash = 10 * 16; // flash built in led after button is pushed.
    }
  }
}
